@@ -11,8 +11,9 @@
 # under the License.
 
 import os
+import warnings
 
-from positional import positional
+from keystoneauth1 import session
 
 
 class Adapter(object):
@@ -24,6 +25,9 @@ class Adapter(object):
     particular client that is using the session. An adapter provides a wrapper
     of client local data around the global session object.
 
+    version, min_version, max_version and default_microversion can all be
+    given either as a string or a tuple.
+
     :param session: The session object to wrap.
     :type session: keystoneauth1.session.Session
     :param str service_type: The default service_type for URL discovery.
@@ -32,7 +36,9 @@ class Adapter(object):
     :param str region_name: The default region_name for URL discovery.
     :param str endpoint_override: Always use this endpoint URL for requests
                                   for this client.
-    :param tuple version: The version that this API targets.
+    :param version: The minimum version restricted to a given Major API.
+                    Mutually exclusive with min_version and max_version.
+                    (optional)
     :param auth: An auth plugin to use instead of the session one.
     :type auth: keystoneauth1.plugin.BaseAuthPlugin
     :param str user_agent: The User-Agent string to set.
@@ -43,13 +49,68 @@ class Adapter(object):
     :param logger: A logging object to use for requests that pass through this
                    adapter.
     :type logger: logging.Logger
+    :param dict allow: Extra filters to pass when discovering API versions.
+                       (optional)
+    :param dict additional_headers: Additional headers that should be attached
+                                    to every request passing through the
+                                    adapter. Headers of the same name specified
+                                    per request will take priority.
+    :param str client_name: The name of the client that created the adapter.
+                            This will be used to create the user_agent.
+    :param str client_version: The version of the client that created the
+                               adapter. This will be used to create the
+                               user_agent.
+    :param bool allow_version_hack: Allow keystoneauth to hack up catalog
+                                    URLS to support older schemes.
+                                    (optional, default True)
+    :param str global_request_id: A global_request_id (in the form of
+                                  ``req-$uuid``) that will be passed on all
+                                  requests. Enables cross project request id
+                                  tracking.
+    :param min_version: The minimum major version of a given API, intended to
+                        be used as the lower bound of a range with
+                        max_version. Mutually exclusive with version.
+                        If min_version is given with no max_version it is as
+                        if max version is 'latest'. (optional)
+    :param max_version: The maximum major version of a given API, intended to
+                        be used as the upper bound of a range with min_version.
+                        Mutually exclusive with version. (optional)
+    :param default_microversion: The default microversion value to send
+                                 with API requests. While microversions are
+                                 a per-request feature, a user may know they
+                                 want to default to sending a specific value.
+                                 (optional)
+    :param int status_code_retries: the maximum number of retries that
+                                    should be attempted for retriable
+                                    HTTP status codes (optional, defaults
+                                    to 0 - never retry).
+    :param list retriable_status_codes: list of HTTP status codes that
+                                        should be retried (optional,
+                                        defaults to HTTP 503, has no effect
+                                        when status_code_retries is 0).
+    :param bool raise_exc: If True, requests returning failing HTTP responses
+                           will raise an exception; if False, the response is
+                           returned. This can be overridden on a per-request
+                           basis via the kwarg of the same name.
     """
 
-    @positional()
+    client_name = None
+    client_version = None
+
     def __init__(self, session, service_type=None, service_name=None,
                  interface=None, region_name=None, endpoint_override=None,
                  version=None, auth=None, user_agent=None,
-                 connect_retries=None, logger=None):
+                 connect_retries=None, logger=None, allow=None,
+                 additional_headers=None, client_name=None,
+                 client_version=None, allow_version_hack=None,
+                 global_request_id=None,
+                 min_version=None, max_version=None,
+                 default_microversion=None, status_code_retries=None,
+                 retriable_status_codes=None, raise_exc=None):
+        if version and (min_version or max_version):
+            raise TypeError(
+                "version is mutually exclusive with min_version and"
+                " max_version")
         # NOTE(jamielennox): when adding new parameters to adapter please also
         # add them to the adapter call in httpclient.HTTPClient.__init__ as
         # well as to load_adapter_from_argparse below if the argument is
@@ -66,6 +127,22 @@ class Adapter(object):
         self.auth = auth
         self.connect_retries = connect_retries
         self.logger = logger
+        self.allow = allow or {}
+        self.additional_headers = additional_headers or {}
+        self.allow_version_hack = allow_version_hack
+        self.min_version = min_version
+        self.max_version = max_version
+        self.default_microversion = default_microversion
+        self.status_code_retries = status_code_retries
+        self.retriable_status_codes = retriable_status_codes
+        self.raise_exc = raise_exc
+
+        self.global_request_id = global_request_id
+
+        if client_name:
+            self.client_name = client_name
+        if client_version:
+            self.client_version = client_version
 
     def _set_endpoint_filter_kwargs(self, kwargs):
         if self.service_type:
@@ -78,6 +155,12 @@ class Adapter(object):
             kwargs.setdefault('region_name', self.region_name)
         if self.version:
             kwargs.setdefault('version', self.version)
+        if self.min_version:
+            kwargs.setdefault('min_version', self.min_version)
+        if self.max_version:
+            kwargs.setdefault('max_version', self.max_version)
+        if self.allow_version_hack is not None:
+            kwargs.setdefault('allow_version_hack', self.allow_version_hack)
 
     def request(self, url, method, **kwargs):
         endpoint_filter = kwargs.setdefault('endpoint_filter', {})
@@ -92,8 +175,40 @@ class Adapter(object):
             kwargs.setdefault('user_agent', self.user_agent)
         if self.connect_retries is not None:
             kwargs.setdefault('connect_retries', self.connect_retries)
+        if self.status_code_retries is not None:
+            kwargs.setdefault('status_code_retries', self.status_code_retries)
+        if self.retriable_status_codes:
+            kwargs.setdefault('retriable_status_codes',
+                              self.retriable_status_codes)
         if self.logger:
             kwargs.setdefault('logger', self.logger)
+        if self.allow:
+            kwargs.setdefault('allow', self.allow)
+        if self.default_microversion is not None:
+            kwargs.setdefault('microversion', self.default_microversion)
+
+        if isinstance(self.session, (session.Session, Adapter)):
+            # these things are unsupported by keystoneclient's session so be
+            # careful with them until everyone has transitioned to ksa.
+            # Allowing adapter allows adapter nesting that auth_token does.
+            if self.client_name:
+                kwargs.setdefault('client_name', self.client_name)
+            if self.client_version:
+                kwargs.setdefault('client_version', self.client_version)
+
+        else:
+            warnings.warn('Using keystoneclient sessions has been deprecated. '
+                          'Please update your software to use keystoneauth1.')
+
+        for k, v in self.additional_headers.items():
+            kwargs.setdefault('headers', {}).setdefault(k, v)
+
+        if self.global_request_id is not None:
+            kwargs.setdefault('headers', {}).setdefault(
+                "X-OpenStack-Request-ID", self.global_request_id)
+
+        if self.raise_exc is not None:
+            kwargs.setdefault('raise_exc', self.raise_exc)
 
         return self.session.request(url, method, **kwargs)
 
@@ -131,6 +246,46 @@ class Adapter(object):
         self._set_endpoint_filter_kwargs(kwargs)
         return self.session.get_endpoint(auth or self.auth, **kwargs)
 
+    def get_endpoint_data(self, auth=None):
+        """Get the endpoint data for this Adapter's endpoint.
+
+        :param auth: The auth plugin to use for token. Overrides the plugin on
+                     the session. (optional)
+        :type auth: keystoneauth1.plugin.BaseAuthPlugin
+
+        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin: if a
+            plugin is not available.
+        :raises TypeError: If arguments are invalid
+
+        :returns: Endpoint data if available or None.
+        :rtype: keystoneauth1.discover.EndpointData
+        """
+        kwargs = {}
+        self._set_endpoint_filter_kwargs(kwargs)
+        if self.endpoint_override:
+            kwargs['endpoint_override'] = self.endpoint_override
+
+        return self.session.get_endpoint_data(auth or self.auth, **kwargs)
+
+    def get_api_major_version(self, auth=None, **kwargs):
+        """Get the major API version as provided by the auth plugin.
+
+        :param auth: The auth plugin to use for token. Overrides the plugin on
+                     the session. (optional)
+        :type auth: keystoneauth1.plugin.BaseAuthPlugin
+
+        :raises keystoneauth1.exceptions.auth_plugins.MissingAuthPlugin: if a
+            plugin is not available.
+
+        :return: The major version of the API of the service discovered.
+        :rtype: tuple or None
+        """
+        self._set_endpoint_filter_kwargs(kwargs)
+        if self.endpoint_override:
+            kwargs['endpoint_override'] = self.endpoint_override
+
+        return self.session.get_api_major_version(auth or self.auth, **kwargs)
+
     def invalidate(self, auth=None):
         """Invalidate an authentication plugin."""
         return self.session.invalidate(auth or self.auth)
@@ -187,15 +342,15 @@ class Adapter(object):
     def delete(self, url, **kwargs):
         return self.request(url, 'DELETE', **kwargs)
 
+    # TODO(efried): Move this to loading.adapter.Adapter
     @classmethod
     def register_argparse_arguments(cls, parser, service_type=None):
-        """Attach arguments to a given argparse Parser for Adapters
+        """Attach arguments to a given argparse Parser for Adapters.
 
         :param parser: The argparse parser to attach options to.
         :type parser: argparse.ArgumentParser
         :param str service_type: Default service_type value. (optional)
         """
-
         adapter_group = parser.add_argument_group(
             'Service Options',
             'Options controlling the specialization of the API'
@@ -237,9 +392,10 @@ class Adapter(object):
             default=os.environ.get('OS_API_VERSION', None),
             help='Which version of the service API to use')
 
+    # TODO(efried): Move this to loading.adapter.Adapter
     @classmethod
     def register_service_argparse_arguments(cls, parser, service_type):
-        """Attach arguments to a given argparse Parser for Adapters
+        """Attach arguments to a given argparse Parser for Adapters.
 
         :param parser: The argparse parser to attach options to.
         :type parser: argparse.ArgumentParser
@@ -338,9 +494,13 @@ class LegacyJsonAdapter(Adapter):
         return resp, body
 
 
+# TODO(efried): Deprecate this in favor of
+#               loading.adapter.register_argparse_arguments
 def register_adapter_argparse_arguments(*args, **kwargs):
     return Adapter.register_argparse_arguments(*args, **kwargs)
 
 
+# TODO(efried): Deprecate this in favor of
+#               loading.adapter.register_service_argparse_arguments
 def register_service_adapter_argparse_arguments(*args, **kwargs):
     return Adapter.register_service_argparse_arguments(*args, **kwargs)
